import os
import unittest
import shutil

from zhihu import ZhihuClient

from .. import db


file_dir = os.path.dirname(os.path.abspath(__file__))
test_dir = os.path.join(file_dir, 'test')
json_path = os.path.join(file_dir, 'test.json')
author = ZhihuClient(json_path).me()
db_path = os.path.join(test_dir, db.author_to_db_filename(author))


class InitDBTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.makedirs(test_dir, exist_ok=True)
        os.chdir(test_dir)

    def tearDown(self):
        self.db.close() if (hasattr(self, 'db') and self.db) else None
        try:
            os.remove(db_path)
        except FileNotFoundError:
            pass

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(test_dir)

    def test_create_db_truely_create_a_file(self):
        self.db = db.create_db(author)
        self.assertTrue(os.path.isfile(db_path))

    def test_create_db_raise_error_when_file_exist(self):
        with open(db_path, 'w') as f:
            f.write('test file')

        with self.assertRaises(FileExistsError):
            self.db = db.create_db(author)

    def test_connect_db_when_file_exist(self):
        # create a db
        self.db = db.create_db(author)
        db.close_db(self.db)

        try:
            self.db = db.connect_db(db_path)
        except FileNotFoundError:
            self.fail("Raise error when try connect a exist database file.")

    def test_connect_db_when_file_not_exist(self):
        with self.assertRaises(FileNotFoundError):
            self.db = db.connect_db(db_path)

    def test_create_table(self):
        self.db = db.create_db(author)
        db.create_table(self.db)
        cursor = self.db.execute(
            """
            select name from sqlite_master where type = 'table';
            """
        )

        self.assertListEqual(list(cursor),
                             [('followers',), ('sqlite_sequence',),
                              ('meta',), ('log',)])

        cursor = self.db.execute(
            """
            select * from followers LIMIT 1;
            """
        )

        row_names = list(map(lambda x: x[0], cursor.description))

        self.assertListEqual(row_names, ['id', 'name', 'in_name'])

    def test_add_one_user_to_db(self):
        self.db = db.create_db(author)
        db.create_table(self.db)
        db.add_user_to_db(self.db, author)

        cursor = self.db.execute(
            """
            SELECT in_name FROM followers;
            """
        )

        for row in cursor:
            self.assertEqual(row[0], author.id)

    def test_is_db_closed_when_closed(self):
        self.db = db.create_db(author)
        self.db.close()
        self.assertTrue(db.is_db_closed(self.db))

    def test_is_db_closed_when_not_closed(self):
        self.db = db.create_db(author)
        self.assertFalse(db.is_db_closed(self.db))

    def test_close_db_when_closed(self):
        self.db = db.create_db(author)
        self.db.close()
        db.close_db(self.db)
        self.assertTrue(db.is_db_closed(self.db))

    def test_close_db_when_not_closed(self):
        self.db = db.create_db(author)
        db.close_db(self.db)
        self.assertTrue(db.is_db_closed(self.db))



import os
import json

file_dir = os.path.dirname(os.path.abspath(__file__))
config_file_path = os.path.join(file_dir, 'config.json')

assert os.path.isfile(config_file_path)

with open(config_file_path, 'r', encoding='utf-8') as f:
    json_string = f.read()
    json_dict = json.loads(json_string)
    for k, v in json_dict.items():
        exec(k + " = '" + v + "'")



import time
import datetime
import logging

from . import daemon
from . import db
from . import common as c
from . import strings as s

from zhihu import ZhihuClient


def calc_message(pattern, me, you, new_follower_num):
    offset = datetime.timedelta(hours=8)
    china_now = datetime.datetime.utcnow() + offset
    my_name = me.name
    follower_num = me.follower_num - new_follower_num
    your_name = you.name

    return pattern.format(now=china_now, my_name=my_name,
                          follower_num=follower_num, your_name=your_name)


class BackendCode(daemon.DaemonProcess):
    def at_exit(self):
        pass

    def run(self, database, msg, interval, log_file, max_old=10):
        c.check_type(database, 'database', str)

        L = logging.getLogger('qqqfome-backend')
        formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s')
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        sh = logging.StreamHandler()
        sh.setLevel(logging.DEBUG)
        sh.setFormatter(formatter)
        L.setLevel(logging.DEBUG)
        L.addHandler(fh)
        L.addHandler(sh)

        try:
            L.info(s.log_connected_to_db.format(database))
            conn = db.connect_db(database)
            L.info(s.success)
        except FileNotFoundError:
            L.exception(s.log_file_not_exist.format(database))
            L.info(s.exit)
            return

        # get cookies from database
        cookies = db.get_cookies(conn)

        if not cookies:
            L.exception(s.log_no_cookies_in_database)
            L.info(s.exit)
            return

        L.info(s.log_get_cookies_from_database)
        L.debug(cookies)

        try:
            client = ZhihuClient(cookies)
            L.info(s.log_build_zhihu_client)
        except Exception as e:
            L.exception(e)
            return

        while True:
            L.info(s.log_start_a_pass)

            i = 0
            while i < 5:
                try:
                    L.info(s.log_build_me)
                    me = client.me()
                    break
                except Exception as e:
                    L.exception(e)
                    i += 1
            else:
                L.error(s.log_fail_to_build_me)
                L.info(s.exit)
                return

            try:
                follower_num = me.follower_num
            except Exception as e:
                L.exception(e)
                L.info(s.log_get_follower_num_failed)
                L.info(s.log_finish_a_pass)
                time.sleep(interval)
                continue

            L.info(s.log_get_follower_num.format(follower_num))
            db.log_to_db(conn, follower_num, s.log_start_a_pass)

            continue_in_db = 0
            new_follower_num = 0

            try:
                for follower in me.followers:
                    L.info(s.log_check_follower.format(
                        follower.name, follower.id))
                    if db.is_in_db(conn, follower.id):
                        L.info(s.log_follower_in_db.format(follower.id))
                        continue_in_db += 1
                    else:
                        L.info(s.log_follower_not_in_db.format(follower.name))
                        continue_in_db = 0

                        L.info(s.log_send_message.format(follower.name))

                        try:
                            message = calc_message(msg, me, follower,
                                                   new_follower_num)
                            new_follower_num += 1
                        except Exception as e:
                            L.exception(e)
                            message = msg

                        L.debug(message)

                        i = 0
                        while i < 5:
                            try:
                                me.send_message(follower, message)
                                break
                            except Exception as e:
                                L.exception(e)
                                L.debug(s.log_send_failed)
                                i += 1
                        else:
                            L.info(s.log_send_pass)
                            continue

                        L.info(s.success)
                        L.info(s.log_add_user_to_db.format(
                            follower.name))
                        db.add_user_to_db(conn, follower)

                    if continue_in_db == max_old:
                        L.info(s.log_continue_reach_max.format(max_old))
                        break
            except Exception as e:
                L.exception(e)

            L.info(s.log_finish_a_pass)
            time.sleep(interval)



import argparse
import logging
import os
import sys

from zhihu import ZhihuClient

from . import common as c
from . import strings as s
from . import db
from . import backend

L = logging.getLogger('qqqfome-entry')


def init_db(cookies):
    if cookies is not None:
        c.check_type(cookies, 'cookies', str)

    # login in with cookies file
    if cookies:
        author = ZhihuClient(cookies=cookies).me()
    # login in terminal
    else:
        client = ZhihuClient()

        try:
            cookies = client.login_in_terminal()
        except KeyboardInterrupt:
            print()
            cookies = ''

        if not cookies:
            L.error(s.log_login_failed)
            L.info(s.exit)
            exit(0)

        author = client.me()

    try:
        conn = db.create_db(author)
        db.create_table(conn)
        db.dump_init_data_to_db(conn, author)
        db.close_db(conn)
        print(s.success)
    except FileExistsError as e:
        L.error(s.file_exist.format(e.filename))
        print(s.failed)


class SetDefaultPID(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)
        default_pid = getattr(namespace, 'pid_file')
        default_log = getattr(namespace, 'log_file')
        try:
            if values:
                filename = os.path.basename(os.path.abspath(values))
                default_pid = default_pid.format(filename)
                default_log = default_log.format(filename)
        except TypeError:
            # pid has already been replaced
            pass
        setattr(namespace, 'pid_file', default_pid)
        setattr(namespace, 'log_file', default_log)


def main():
    parser = argparse.ArgumentParser(prog='qqqfome',
                                     description='Thank-you-follow-me cli.')

    parser.add_argument('-v', '--verbose', dest='verbose',
                        action='count', default=0,
                        help=s.cmd_help_print_info)
    parser.add_argument('-c', '--cookies', dest='cookies', metavar='FILE',
                        help=s.cmd_help_cookies,
                        type=str)
    parser.add_argument('-p', '--pid-file', dest='pid_file', metavar='FILE',
                        help=s.cmd_help_pid_file,
                        type=str, default='{0}.pid')
    parser.add_argument('-l', '--log-file', dest='log_file', metavar='FILE',
                        help=s.cmd_help_log_file,
                        type=str, default='{0}.log')
    parser.add_argument('-t', '--time', dest='time', metavar='INTERVAL',
                        help=s.cmd_help_time,
                        type=int, default=90)

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-m', '--message', dest='message',
                       help=s.cmd_help_message,
                       type=str, default=s.default_message)
    group.add_argument('-M', '--message-file', dest='message_file',
                       metavar='FILE',
                       help=s.cmd_help_message,
                       type=str)

    parser.add_argument('-s', '--stop-at', dest='stop_at', metavar='NUM',
                        help=s.cmd_help_stop_at, type=int, default=10)
    parser.add_argument('-d', '--daemon', dest='daemon', action='store_true',
                        default=False, help='work in daemon mode')
    parser.add_argument('command', help=s.cmd_help_command, type=str,
                        choices=['init', 'start', 'stop'])
    parser.add_argument('file', help=s.cmd_help_file, type=str,
                        action=SetDefaultPID, nargs='?')

    args = parser.parse_args()

    # Logger settings
    level = logging.ERROR
    if args.verbose == 1:
        level = logging.INFO
    if args.verbose >= 2:
        level = logging.DEBUG

    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    db.set_logger_level(level)
    db.set_logger_handle(ch)

    L.setLevel(level)
    L.addHandler(ch)

    L.debug(args)

    if args.command == "init":
        init_db(args.cookies)
    elif args.file is None:
            parser.error(s.cmd_help_no_file_error.format(args.command))

    if args.command == 'start':
        if args.message_file is not None:
            with open(args.message_file, 'r') as f:
                args.message = f.read()
        if args.daemon:
            p = backend.BackendCode(args.pid_file)
            try:
                p.start(args.file, args.message, args.time, args.log_file)
            except OSError:     # daemon mode not support the system
                p.run(args.file, args.message, args.time, args.log_file)
        else:
            p = backend.BackendCode(args.pid_file, stdin=sys.stdin,
                                    stdout=sys.stdout, stderr=sys.stderr)
            p.run(args.file, args.message, args.time, args.log_file)
    elif args.command == 'stop':
        try:
            p = backend.BackendCode(args.pid_file)
            p.stop()
        except FileNotFoundError as e:
            L.error(e)


if __name__ == '__main__':
    main()



import warnings


def ignore_warnings(test_func):
    def do_test(self, *args, **kwargs):
        warnings.filterwarnings("ignore", category=ResourceWarning)
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        test_func(self, *args, **kwargs)
    return do_test



from . import strings as s


def check_type(var, name, t):
    if not isinstance(t, type(str)):
        t = type(t)
    if not isinstance(var, t):
        raise ValueError(s.type_error.format(
            name, str(t), str(type(var))))


import os
import sqlite3
import json
import logging
import datetime

from zhihu import Author, ZhihuClient

from . import common as c
from . import strings as s

L = logging.getLogger('qqqufome-db')


def set_logger_level(level):
    c.check_type(level, 'level', logging.NOTSET)
    global L
    L.setLevel(level)


def set_logger_handle(handle):
    L.addHandler(handle)


def author_to_db_filename(author):
    c.check_type(author, 'author', Author)

    return author.id + '.sqlite3'


def create_db(author):
    c.check_type(author, 'author', Author)

    filename = author_to_db_filename(author)

    L.info(s.log_get_user_id.format(filename))

    if os.path.isfile(filename):
        e = FileExistsError()
        e.filename = filename
        raise e

    L.info(s.log_db_not_exist_create.format(filename))

    db = sqlite3.connect(author_to_db_filename(author))

    L.info(s.log_connected_to_db.format(filename))

    return db


def connect_db(database):
    c.check_type(database, 'database', str)

    if not os.path.isfile(database):
        e = FileNotFoundError()
        e.filename = database
        raise e

    return sqlite3.connect(database)


def create_table(db: sqlite3.Connection):
    c.check_type(db, 'db', sqlite3.Connection)

    L.info(s.log_create_table_in_db)
    with db:
        db.execute(
            '''
           CREATE TABLE followers
           (
           id       INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
           name     TEXT            NOT NULL,
           in_name  TEXT            NOT NULL
           );
           '''
        )

        db.execute(
            """
           CREATE TABLE meta
           (
           id       INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
           name     TEXT            NOT NULL,
           in_name  TEXT            NOT NULL,
           cookies  TEXT            NOT NULL
           );
            """
        )

        db.execute(
            """
           CREATE TABLE log
           (
           id               INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
           time             DATETIME        NOT NULL,
           follower_number  INT             NOT NULL,
           increase         INT             NOT NULL,
           message          TEXT            NOT NULL
           );
            """
        )
    L.info(s.success)


def add_user_to_db(db, author):
    c.check_type(db, 'db', sqlite3.Connection)
    c.check_type(author, 'author', Author)

    with db:
        L.debug(s.log_add_user_to_db.format(author.name))
        db.execute(
            """
            INSERT INTO followers
            (name, in_name) VALUES
            ( ?,      ?   );
            """,
            (author.name, author.id)
        )


def dump_init_data_to_db(db, author):
    c.check_type(db, 'db', sqlite3.Connection)
    c.check_type(author, 'author', Author)

    # meta data
    with db:
        name = author.name
        in_name = author.id
        cookies = json.dumps(author._session.cookies.get_dict())

        db.execute(
            """
            INSERT INTO meta
            (name,    in_name,   cookies) VALUES
            (  ?,        ?,         ?   );
            """,
            (name, in_name, cookies)
        )

    # followers
    L.info(s.log_start_get_followers.format(author.name))
    with db:
        for _, follower in zip(range(100), author.followers):
            add_user_to_db(db, follower)

    # log
    with db:
        log_to_db(db, author.follower_num, s.log_db_init)


def is_db_closed(db):
    c.check_type(db, 'db', sqlite3.Connection)

    try:
        with db:
            db.execute(
                """
                SELECT name from sqlite_master where type = 'table';
                """
            )
        return False
    except sqlite3.ProgrammingError:
        return True


def close_db(db):
    c.check_type(db, 'db', sqlite3.Connection)

    if not is_db_closed(db):
        db.close()
        L.info(s.log_close_db)


def get_cookies(db):
    c.check_type(db, 'db', sqlite3.Connection)

    cursor = db.execute('SELECT cookies from meta')

    row = cursor.fetchone()

    if row is None:
        return None

    return row[0]


def log_to_db(db, follower_num, message):
    c.check_type(db, 'db', sqlite3.Connection)
    c.check_type(follower_num, 'follower_num', int)
    c.check_type(message, 'message', str)

    cursor = db.execute(
        """
        SELECT follower_number FROM log ORDER BY id DESC;
        """
    )

    row = cursor.fetchone()

    if row:
        increase = follower_num - row[0]
    else:
        # first log
        increase = 0

    with db:
        db.execute(
            """
            INSERT INTO log
            (time, follower_number, increase, message) VALUES
            ( ?,           ?,           ?,       ?   );
            """,
            (datetime.datetime.now(), follower_num, increase, message)
        )


def is_in_db(db, in_name):
    c.check_type(db, 'db', sqlite3.Connection)
    c.check_type(in_name, 'in_name', str)

    with db:
        cursor = db.execute(
            """
            SELECT * FROM followers WHERE in_name = ?;
            """,
            (in_name,)
        )

        row = cursor.fetchone()

        return row is not None



"""
Make some code running in daemon.

Modify from Open Source file:
    https://github.com/serverdensity/python-daemon/blob/master/daemon.py
"""

import os
import sys
import signal
import atexit
import time

from . import common as c
from . import strings as s


class DaemonProcess:
    def __init__(self, pidfile,
                 stdin=os.devnull, stdout=os.devnull, stderr=os.devnull,
                 workdir='.', umask='022'):
        c.check_type(pidfile, 'pidfile', str)
        c.check_type(workdir, 'workdir', str)
        c.check_type(umask, 'umask', str)

        self._pidfile = pidfile
        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr
        self._work_dir = workdir
        self._alive = True

        try:
            self.umask = int(umask, base=8)
        except ValueError as e:
            raise ValueError(s.daemon_umask_error)

    def run(self, *args, **kwargs):
        """
        Override me.
        """
        pass

    def at_exit(self):
        """
        Override me.
        """
        pass

    def _write_pid_file(self, pid):
        with open(self._pidfile, 'w', encoding='utf-8') as f:
            f.writelines([str(pid)])

    def _del_pid_file(self):
        self.at_exit()
        os.remove(self._pidfile)

    def get_pid_from_pidfile(self):
        try:
            with open(self._pidfile, 'r', encoding='utf-8') as f:
                return int(f.read().strip())
        except IOError:
            return None
        except SystemExit:
            return None

    def _make_me_daemon(self):
        # first fork
        try:
            pid = os.fork()
            if pid != 0:
                sys.exit(0)
        except OSError as e:
            raise OSError(s.daemon_fork_error.format(e.errno, e.strerror))

        os.setsid()
        os.umask(self.umask)
        os.chdir(self._work_dir)

        # second fork
        try:
            pid = os.fork()
            if pid != 0:
                sys.exit(0)
        except OSError as e:
            raise OSError(s.daemon_fork_error.format(e.errno, e.strerror))

        # redirect streams
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(self._stdin, 'r') if isinstance(self._stdin, str) \
            else self._stdin
        so = open(self._stdout, 'a+') if isinstance(self._stdout, str) \
            else self._stdout
        if self._stderr:
            se = open(self._stderr, 'a+') if isinstance(self._stderr, str) \
                else self._stdout
        else:
            se = so
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # set signal handler
        def signal_handler(signum, frame):
            self._alive = False
            exit(0)

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        pid = os.getpid()
        print(pid)
        atexit.register(self._del_pid_file)
        self._write_pid_file(pid)

    def start_when_pid_file_exist(self):
        raise FileExistsError(
            s.daemon_pid_file_exist_error.format(self._pidfile))

    def start(self, *args, **kwargs):
        if not (sys.platform.startswith('linux') or sys.platform.startswith(
                'darwin')):
            raise OSError(s.daemon_os_error)

        pid = self.get_pid_from_pidfile()

        if pid is not None:
            self.start_when_pid_file_exist()

        self._make_me_daemon()
        self.run(*args, **kwargs)

    def stop_when_pid_file_not_exist(self):
        raise FileNotFoundError(
            s.daemon_pid_file_not_found_error.format(self._pidfile))

    def stop(self):
        pid = self.get_pid_from_pidfile()

        if pid is None:
            # Just to be sure. A ValueError might occur if the PID file is
            # empty but does actually exist
            if os.path.exists(self._pidfile):
                os.remove(self._pidfile)

            self.stop_when_pid_file_not_exist()
        else:
            try:
                i = 0
                while 1:
                    os.kill(pid, signal.SIGTERM)
                    time.sleep(0.1)
                    i += 1
                    if i % 10 == 0:
                        os.kill(pid, signal.SIGHUP)
            except OSError as err:
                err = str(err)
                if err.find("No such process") > 0:
                    if os.path.exists(self._pidfile):
                        os.remove(self._pidfile)
                else:
                    raise OSError(s.daemon_can_not_kill_process.format(pid))

    def is_running(self):
        pid = self.get_pid_from_pidfile()

        return self._pid and os.path.exists('/proc/%d' % pid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import ast

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def extract_version():
    with open('qqqfome/__init__.py', 'rb') as f_version:
        ast_tree = re.search(
            r'__version__ = (.*)',
            f_version.read().decode('utf-8')
        ).group(1)
        if ast_tree is None:
            raise RuntimeError('Cannot find version information')
        return str(ast.literal_eval(ast_tree))


packages = ['qqqfome', 'qqqfome.test']
package_data = {'qqqfome': ['config.json']}

version = extract_version()

long_description = ''

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='qqqfome',
    version=version,
    keywords=['internet', 'daemon', 'sqlite'],
    description='I\'m a daemon server that '
                'auto send message to your zhihu new followers.',
    long_description=long_description,
    author='7sDream',
    author_email='didislover@gmail.com',
    license='MIT',
    url='https://github.com/7sDream/qqqfome',

    install_requires=[
        'zhihu-py3', 
    ],
    
    packages=packages,
    package_data=package_data,

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: No Input/Output (Daemon)',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Database', 
        'Topic :: Internet :: WWW/HTTP'
    ],

    entry_points={
        'console_scripts': [
            'qqqfome = qqqfome.entry:main'
        ]
    }
)



import unittest

from .. import common as c


class CommonFuncTest(unittest.TestCase):
    def test_check_type_success_with_ref_is_a_value(self):
        try:
            c.check_type(1, "num", 1)
            a = 1
            c.check_type(1, "num", a)
        except ValueError:
            self.fail("check_type function raise a exception "
                      "when the check should be success.")

    def test_check_type_success_with_ref_is_a_type(self):
        try:
            c.check_type(1, "num", int)
        except ValueError:
            self.fail("check_type function raise a exception "
                      "when the check should be success.")

    def test_check_type_fail_with_ref_is_a_value(self):
        with self.assertRaises(ValueError):
            c.check_type(1, "string", "")
        with self.assertRaises(ValueError):
            string = ""
            c.check_type(1, 'string', string)

    def test_check_type_fail_with_ref_is_a_type(self):
        with self.assertRaises(ValueError):
            c.check_type(1, "string", str)
